﻿// AS2.0 Back Propagation Neural Net

// This code is translated from the java found at the following url:
// <http://www.patol.com/java/NN/index.html>
// I've tried contacting the guy and no luck (the code is over 10 years old!)
// - but I've done a fair bit of research on neural nets so I've got a rough idea of how it works.
// All in all - Patol's code here is pretty fast and stable and open source and the XML
// features are all mine, as well as Flash support.

// This is a 3-layer Back-Propagation Neural Net.
// Think of it as a child's shape sorter - you train it on certain shapes
// and it will tell you according to the data you give it what shape it thinks the
// sample data is most like what it knows.

// It learns by changing weights on it's internal connections. Thus the more you tell
// it to recognise a square the more fussy it gets.

// Picking the right training set will ensure the BPN does not get confused and
// allow you to identify patterns from noisy data like handwriting, sound and images
// The only advice I can offer is to experiment as much as possible.

// Aaron Steed 2007 - st3ed@hotmail.com

class BPN{
	var FIRE:Number = 0.999;
	var NEUTRAL:Number = 0.0;
	var DOWN:Number = -0.999;
	var INIT:Number = -1.0;
	var DELTA:Number = 0.0;
	var OUTPUT:Number = 1.0;
	var instantError:Number;
	var totalError:Number;
	var absoluteError:Number;
	var inpA:Array;		 // activations
	var hidA:Array;		 // activations
	var hidN:Array;		 // sum of products
	var hidD:Array;		 // output error
	var hidW:Array;		 // connection weights
	var outA:Array;		 // activations
	var outN:Array;		 // sum of products
	var outD:Array;		 // output error

	var oldD:Array;		 // old output error
	var outW:Array;		 // connection weights
	var Ninp:Number;	 // number of neurons on input  layer
	var Nhid:Number;	 // number of neurons on hidden layer
	var Nout:Number;	 // number of neurons on output layer
	var eida:Number;	 // learning rate
	var theta:Number;	 // sigmoid thresold
	var elast:Number;	 // elastics of sigmoid
	var momentum:Number;
	// Constructor
	function BPN(Ninp:Number, Nhid:Number, Nout:Number, eida:Number, theta:Number, elast:Number, momentum:Number){		
		this.Ninp = Ninp;
		this.Nhid = Nhid;
		this.Nout = Nout;
		this.inpA = new Array();// [Ninp]
		this.hidW = new Array();// [Nhid][Ninp]
		this.hidA = new Array();// [Nhid]
		this.hidN = new Array();// [Nhid]
		this.hidD = new Array();// [Nhid]
		this.outW = new Array();// [Nout][Nhid]
		this.outA = new Array();// [Nout]
		this.outN = new Array();// [Nout]
		this.outD = new Array();// [Nout]
		this.oldD = new Array();// [Nout]
		this.eida = eida;
		this.theta = theta;
		this.elast = elast;
		this.momentum = momentum;
		this.init();
	}
	// Sigmoid transformation (smooths out the signal from a slope to a nice S shaped curve)
	private function sigmoid(x:Number):Number{
		// sig = 1.0 - Math.exp(-1.5* x + theta);
		var sig = ( 1.0 / (1.0 + Math.exp(-1.0 * elast * x + theta)) * 2.0 - 1.0);
		return sig; // (sig < -1.0)? (-1.0):(sig> 1.0)? (1.0): sig;
	}
	// Unsure what this does as yet - obviously a sigmoid transformation of a sort
	private function d1sigmoid(x:Number):Number{
		//double sig = sigmoid(n,x);
		return 2.0 * Math.exp(-1.0 * elast * x -  theta)/(1+Math.exp(-2.0 * elast * x -  theta));
	}
	// Push the current data through the network and alter the weights
	function feedForward():Void{
		var sum2 = 0.0;
		for(var i = 0; i < Nhid; i++){
			for(var j = 0; j < Ninp; j++) sum2 += hidW[i][j] * inpA[j];
			hidN[i] = sum2;
			hidA[i] = sigmoid(sum2);
		}
		for(var i = 0; i < Nout; i++){
			sum2 = 0.0;
			for(var j = 0; j < Nhid; j++) sum2 += outW[i][j] * hidA[j];
			outN[i] = sum2;
		}
	}
	// Calculate the error
	function computeDelta(m:Number):Number{
		outD[m] = (outA[m] - sigmoid(outN[m])) * (d1sigmoid(outN[m]) + 0.1);
		for(var i = 0; i < Nhid; i++) outW[m][i] += outD[m] * hidA[i] * eida;
		return outD[m]; // /(sig1(n,n->outN[m])+0.1);
	}
	// What it says on the tin
	function updateWeights():Void{
		var sum2 = 0.0;
		for(var m = 0; m < Nhid; m++){
			for(var i = 0; i < Nout; i++){
				sum2 += outD[i] * outW[i][m];
			}
			sum2 *= d1sigmoid(hidN[m]);
			for(var i = 0; i < Ninp; i++) hidW[m][i] += eida * sum2 * inpA[i];
		}
	}
	// Shorthand for a random number method
	function frandom(min:Number, max:Number):Number{
		return Math.random()*(max - min) + min;
	}
	// Put an array of data into the in-ports of the neural net
	// The data gets filtered through the wieghts and the outputs can be read as a percentage likelyhood
	// of a given identity
	function propagate(vector:Array):Void{
		var	sum2 = 0.0;
		if(vector.length != Ninp) trace("Error: Vector size don't match Network input size !");
		for(var i = 0; i < Ninp; i++) inpA[i] = vector[i];
			for(var i = 0; i < Nhid; i++){
			for(var j = 0; j < Ninp; j++) sum2 += hidW[i][j] * inpA[j];
			hidA[i] = sigmoid(sum2);
		}
		for(var i = 0; i < Nout; i++){
			sum2 = 0.0;
			for(var j = 0; j < Nhid; j++) sum2 += outW[i][j] * hidA[j];
			outA[i] = sigmoid(sum2);
		}
	}
	// Initialise the arrays and randomise values
	function init():Void{
		for(var i = 0; i < Ninp; i++) inpA[i] = frandom(-1.0,1.0);
		for(var i = 0; i < Nhid;i++){
			hidA[i] = frandom(-1.0,1.0);
			hidW[i] = new Array();
			for(var m = 0; m < Ninp; m++) hidW[i][m] = frandom(-1.0,1.0);
		}
		for(var i = 0; i < Nout; i++){
			outW[i] = new Array();
			for(var m = 0; m < Nhid; m++) outW[i][m] = frandom(-1.0,1.0);
		}
		totalError = 0.0;
		absoluteError = 0.0;
	}
	// I've yet to find a use for this function - it really doesn't help much in training
	function trickForLearning(min:Number, max:Number):Void{
		var i;
		for(i = 0; i < Ninp; i++) inpA[i] += frandom(min,max);
		inpA[i] = (inpA[i] > FIRE)? FIRE: (inpA[i] < DOWN) ? DOWN : inpA[i];
	}
	// Execute a training epoch
	// For a given input you expect a given output.
	// Say - you have an array describing a shape and you want outA[3] to register 0.99
	// and all the other outA values to register 0.0 when you propagate this shape through the
	// array.
	function learnVector(_in:Array, out:Array):Void{
		if(_in.length != Ninp) trace("Error: In Vector size don't match Network input size !");
		if(out.length != Nout) trace("Error: Out Vector size don't match Network output size !");
		for(var i = 0; i < Ninp; i++) inpA[i] = _in[i];
		for(var i = 0; i < Nout; i++) outA[i] = out[i];
		this.feedForward();
		totalError = 0.0;
		absoluteError = 0.0;
		for(var j = 0; j < Nout ; j++){
			instantError = computeDelta(j);
			totalError += instantError;
			absoluteError += Math.abs(instantError);
		}
		updateWeights();
		eida *= momentum;
	}
	// Return an XML object describing the BPN
	function getXML(){
		var xml = new XML();
		var root = xml.createElement("root");
		var bpnNode = xml.createElement("bpn");
		bpnNode.attributes.Ninp = Ninp;
		bpnNode.attributes.Nhid = Nhid;
		bpnNode.attributes.Nout = Nout;
		bpnNode.attributes.eida = eida;
		bpnNode.attributes.theta = theta;
		bpnNode.attributes.elast = elast;
		bpnNode.attributes.momentum = momentum;
		root.appendChild(bpnNode);
		var weights = xml.createElement("weights");
		for(var i = 0; i < Ninp; i++){
			for(var j = 0; j < Nhid; j++){
				var weightNode = xml.createElement("hidW");
				weightNode.attributes.j = j;
				weightNode.attributes.i = i;
				weightNode.attributes.value = hidW[j][i];
				weights.appendChild(weightNode);
			}
		}
		for(var i = 0; i < Nhid; i++){
			for(var j = 0; j < Nout; j++){
				var weightNode = xml.createElement("outW");
				weightNode.attributes.j = j;
				weightNode.attributes.i = i;
				weightNode.attributes.value = outW[j][i];
				weights.appendChild(weightNode);
			}
		}
		root.appendChild(weights);
		xml.appendChild(root);
		return xml;
	}
	// Load a saved state of the BPN
	function loadXML(path:String){
		var xml = new XML();
		xml["bpn"] = this;
		xml.onLoad = function(success){
			if(success){
				this["bpn"].decodeXML(this);
			} else {
				trace("failed to load XML");
			}
		}
		xml.load(path);
	}
	// Extrapolate a saved state from an XML
	function decodeXML(xml:XML){
		var rootNode = xml.firstChild;
		if(rootNode.nodeName == "root"){
			var myNode = rootNode.firstChild;
			while(myNode != null){
				if(myNode.nodeName == "bpn"){
					Ninp = parseInt(myNode.attributes.Ninp);
					Nhid = parseInt(myNode.attributes.Nhid);
					Nout = parseInt(myNode.attributes.Nout);
					eida = parseFloat(myNode.attributes.eida);
					theta = parseFloat(myNode.attributes.theta);
					elast = parseFloat(myNode.attributes.elast);
					momentum = parseFloat(myNode.attributes.momentum);
					init();
				}
				if(myNode.nodeName == "weights"){
					var weightNode = myNode.firstChild;
					this.inpA = new Array();// [Ninp]
					this.hidW = new Array();// [Nhid][Ninp]
					this.hidA = new Array();// [Nhid]
					this.hidN = new Array();// [Nhid]
					this.hidD = new Array();// [Nhid]
					this.outW = new Array();// [Nout][Nhid]
					this.outA = new Array();// [Nout]
					this.outN = new Array();// [Nout]
					this.outD = new Array();// [Nout]
					this.oldD = new Array();// [Nout]
					for(var i = 0; i < Nhid; i++){
						hidW.push(new Array());
						for(var j = 0; j < Ninp; j++)
							hidW[i].push(0);
					}
					for(var i = 0; i < Nout; i++){
						outW.push(new Array());
						for(var j = 0; j < Nhid; j++)
							outW[i].push(0);
					}
					while(weightNode != null){
						if(weightNode.nodeName == "hidW"){
							var i = weightNode.attributes.i;
							var j = weightNode.attributes.j;
							hidW[j][i] = parseFloat(weightNode.attributes.value);
						}
						if(weightNode.nodeName == "outW"){
							var i = weightNode.attributes.i;
							var j = weightNode.attributes.j;
							outW[j][i] = parseFloat(weightNode.attributes.value);
						}
						weightNode = weightNode.nextSibling;
					}
				}
				myNode = myNode.nextSibling;
			}
		}
	}
}