﻿// Freeline drawing class
// allows creation of a line of points that can be trimmed back to a specific size for analysis

// This code also has a line intersection test so you can use various methods to detect gestures

// Should you want to detect a straight line you can use the method I use - detecting variance:

/*
var angles = line.getAngles();
// are all the angles the same?
// is it a line?
var variance = 0;
for(var i = 1; i < angles.length; i++){
	variance += Math.abs(angles[i] - angles[i-1]);
}
if(variance < 0.7){
	return true;
}
*/

class FreeLine{
	var p:Array;
	var start:Object;
	var end:Object;
	// Constructor
	function FreeLine(){
		p = new Array();
		start = new Object();
		end = new Object();
	}
	// Return a duplicate of this object
	function copy():FreeLine{
		var fl = new FreeLine();
		for(var i = 0; i < p.length; i++){
			fl.p[i] = {x:p[i].x, y:p[i].y};
		}
		return fl;
	}
	// Add a point to the line
	function add(xy:Object):Void{
		p.push(xy);
		if(p.length == 1){
			start = {x:p[0].x, y:p[0].y};
		} else {
			end = {x:p[p.length-1].x, y:p[p.length-1].y};
		}
	}
	// Reduce the length of the line to a specific value
	function simplify(len:Number):Void{
		if(p.length == 0) return;
		// Ensure the average distance between points is similar enough for the BPN
		var tolerance = 6;
		var leastDistance = Number.MAX_VALUE;
		for(var i = 1; i < p.length; i++){
			var dist = (p[i].x - p[i-1].x) * (p[i].x - p[i-1].x) + (p[i].y - p[i-1].y) * (p[i].y - p[i-1].y);
			if(dist < leastDistance && dist > tolerance){
				leastDistance = dist;
			}
		}
		while(true){
			var count = 0;
			for(var i = 1; i < p.length; i++){
				var dist = (p[i].x - p[i-1].x) * (p[i].x - p[i-1].x) + (p[i].y - p[i-1].y) * (p[i].y - p[i-1].y);
				if(dist > leastDistance){
					p.splice(i, 0, {x:(p[i].x+p[i-1].x)*0.5, y:(p[i].y+p[i-1].y)*0.5});
					count++;
				}
			}
			if(count == 0) break;
		}
		// If path is too short - double it up by interpolating points
		while(p.length < len){
			var np = new Array();
			np.push(p[0]);
			for(var i = 1; i < p.length; i++){
				np.push({x:(p[i].x+p[i-1].x)*0.5, y:(p[i].y+p[i-1].y)*0.5});
				np.push(p[i]);
			}
			p = np;
		}
		// Path should now be too long - take "len" amount of steps along the array taking points
		if(p.length > len){
			var step = p.length / len;
			var j = 0;
			var np = new Array();
			for(var i = 0; i < len; i++){
				np.push(p[Math.floor(j)]);
				j += step;
			}
			p = np;
		}
	}
	// Draw the line on a MovieClip
	function draw(mc:MovieClip, col:Number):Void{
		mc.clear();
		mc.lineStyle(4, col);
		mc.moveTo(p[0].x, p[0].y);
		for(var i = 1; i < p.length; i++){
			mc.lineTo(p[i].x, p[i].y);
		}
	}
	// Draw filled shape on a MovieClip
	function drawFilled(mc:MovieClip, fill:Number):Void{
		mc.clear();
		mc.lineStyle(2, 0);
		mc.beginFill(fill);
		mc.moveTo(p[0].x, p[0].y);
		for(var i = 1; i < p.length; i++){
			mc.lineTo(p[i].x, p[i].y);
		}
	}
	// Return an array of the angles between all of the points
	// Note: this array will have a length of p.length-1
	function getAngles():Array{
		var angles = new Array();
		var angle = 0.0;
		for(var i = 1; i < p.length; i++){
			var theta = Math.atan2(p[i].y - p[i-1].y, p[i].x - p[i-1].x) / Math.PI;
			angles.push(theta);
			angle = theta;
		}
		return angles;
	}
	// Test if there is an overlap in the line - note a single overlap will return 2, double 4 and so on
	// I've updated this to my new Vector class and this is a techinique I added for a gesture game
	// so it's untested in this open-source stuff
	function overlap():Number{
		if(p.length == 0) return 0;
		var v = new Array();
		// push the original start position in to compensate for simplification
		v.push(new Vector(start, p[0]));
		for(var i = 1; i < p.length; i++){
			v.push(new Vector(p[i-1], p[i]));
		}
		// push the original end position in to compensate for simplification
		v.push(new Vector(p[p.length-1], end));
		// count the intersections
		var count = 0;
		for(var i = 0; i < v.length; i++){
			for(var j = 0; j < v.length; j++){
				var ip0 = Vector.intersectionPoint(v[i], v[j]);
				var ip1 = Vector.intersectionPoint(v[j], v[i]);
				if(ip0.t > 0 && ip0.t < 1 && ip1.t > 0 && ip1.t < 1){
					count++;
				}
			}
		}
		return count;
	}
}