﻿// Gesture Recognition test - Aaron Steed 2007

// Updated July 2007 - better freeline object, random order training and line intersection methods

class GT{
	var stage:MovieClip;
	var canvas:MovieClip;
	var graph:MovieClip;
	var loop:Number;
	var frameRate = 30;
	var bpn:BPN;
	var pathLength = 21;
	var sampleNum = 5;
	var mousePressed;
	var mouseReleased;
	var line:FreeLine;
	var sampleSet:Array;
	var outs:Array;
	var cyclicTraining:Boolean;
	var weakestTraining:Boolean;
	var epochs:Number = 0;
	// Constructor
	function GT(mc:MovieClip){
		this.stage = mc;
		//inputs, hidden, outputs, learning, sigmoid threshold, sigmoid elasticity, momentum
		bpn = new BPN(pathLength-1, Math.round((pathLength + sampleNum) / 2), sampleNum, 0.05, 0.5, 0.5, 0.999);
		canvas = stage.canvas;
		graph = stage.graph;
		canvas["g"] = this;
		canvas.onPress = function(){
			this["g"].mousePressed = true;
			this["g"].line = new FreeLine();
		}
		canvas.onRelease = function(){
			this["g"].mousePressed = false;
			this["g"].mouseReleased = true;
		}
		line = new FreeLine();
		sampleSet = new Array();
		outs = new Array();
		for(var i = 0; i < sampleNum; i++){
			outs[i] = new Array();
			for(var j = 0; j < sampleNum; j++){
				outs[i][j] = i == j ? 0.99 : 0.0;
			}
		}
		bpn.loadXML("http://www.robotacid.com/flash/gesture/network.xml");
		loadSampleSetXML("http://www.robotacid.com/flash/gesture/samples.xml");
		loop = setInterval(this, "main", frameRate);
	}
	// Main program loop
	function main(){
		if(mousePressed){
			line.add({x:stage._xmouse, y:stage._ymouse});
		}
		if(mouseReleased){
			line.simplify(pathLength);
			mouseReleased = false;
		}
		if(cyclicTraining){
			cyclicTrainingEpoch();
		}
		if(weakestTraining){
			cyclicTrainingEpoch();
		}
		line.draw(canvas.holder);
	}
	// Get the samples from XML and add them to the sampleSet Array
	// plus draw them out so we can see what samples we've got
	function decodeSamplesXML(xml:XML){
		var rootNode = xml.firstChild;
		var sampleNode = rootNode.firstChild;
		sampleSet = new Array();
		while(sampleNode != null){
			if(sampleNode.nodeName == "sample"){
				sampleSet.push(new Array());
				var angleNode = sampleNode.firstChild;
				while(angleNode != null){
					if(angleNode.nodeName == "a"){
						sampleSet[sampleSet.length-1].push(parseFloat(angleNode.attributes.value));
					}
					angleNode = angleNode.nextSibling;
				}
			}
			sampleNode = sampleNode.nextSibling;
		}
	}
	// Re-initialise the neural net for a round of training
	function resetNetwork(){
		bpn = new BPN(pathLength-1, Math.round((pathLength + sampleNum) / 2), sampleNum, 0.05, 0.5, 0.5, 0.999);
		epochs = 0;
		graph.clear();
	}
	// Train the net on each sample in a random order
	// Note: there are other ways of training - but this seems to work most of the time
	// Randomised version of cyclic training
	function cyclicTrainingEpoch(){
		graph.lineStyle(1, 0, 20);
		var pick = randomIndex(sampleNum);
		for(var i = 0; i < sampleNum; i++){
			var n = pick[i];
			bpn.learnVector(sampleSet[n], outs[n]);
			graph.moveTo(epochs, 0);
			graph.lineTo(epochs, 100 * bpn.outD[n]);
			stage["s"+n].text = n+":"+bpn.outD[n];
		}
		epochs++;
	}
	// Get the neural net to identify a drawing made by the user
	function classify(){
		var temp = line.getAngles();
		if(temp.length != 0){
			bpn.propagate(temp);
		}
		var bestMatch = -1;
		var bestMatchScore = -1.0;
		var output = "";
		for(var i = 0; i < bpn.outA.length; i++){
			output += bpn.outA[i]+" ";
			if(bpn.outA[i] > bestMatchScore){
				bestMatch = i;
				bestMatchScore = bpn.outA[i];
			}
		}
		//trace(output);
		//trace("best match:"+bestMatch);
		stage.info.text = "best match: "+bestMatch;
	}
	// Save the current drawing on screen to a sampleSet and draw it on the swatch button
	function applyToSwatch(num:Number){
		var mc = stage["swatch_"+num].holder;
		mc.clear();
		mc.lineStyle(1,0);
		mc.moveTo(0,0);
		var angles = line.getAngles();
			if(angles.length != 0){
			var x = 0;
			var y = 0;
			var theta = 0.0;
			for(var i = 0; i < angles.length; i++){
				theta = Math.PI * angles[i];
				x += Math.cos(theta) * 3;
				y += Math.sin(theta) * 3;
				mc.lineTo(x, y);
			}
			sampleSet[num] = angles;
		}
	}
	// Create an XML object representing the sampleSet and print it
	function getSampleSetXML(){
		var xml = new XML();
		var rootNode = xml.createElement("root");
		for(var i = 0; i < sampleSet.length; i++){
			var sampleNode = xml.createElement("sample");
			sampleNode.attributes.num = i;
			for(var j = 0; j < sampleSet[i].length; j++){
				var angleNode = xml.createElement("a");
				angleNode.attributes.value = sampleSet[i][j];
				sampleNode.appendChild(angleNode);
			}
			rootNode.appendChild(sampleNode);
		}
		xml.appendChild(rootNode);
		return xml;
	}
	// Load an XML with a sample set in it
	function loadSampleSetXML(path:String){
		var xml = new XML();
		xml["g"] = this;
		xml.onLoad = function(success){
			if(success){
				this["g"].decodeSampleSetXML(this);
			} else {
				trace("failed to load XML");
			}
		}
		xml.load(path);
	}
	// Fill sampleSet array with values from XML and draw swatches
	function decodeSampleSetXML(xml:XML){
		var rootNode = xml.firstChild;
		if(rootNode.nodeName == "root"){
			var sampleNode = rootNode.firstChild;
			sampleSet = new Array();
			while(sampleNode != null){
				if(sampleNode.nodeName == "sample"){
					var angleNode = sampleNode.firstChild;
					sampleSet.push(new Array());
					while(angleNode != null){
						if(angleNode.nodeName == "a"){
							var value = angleNode.attributes.value;
							sampleSet[sampleSet.length-1].push(parseFloat(value));
						}
						angleNode = angleNode.nextSibling;
					}
				}
				sampleNode = sampleNode.nextSibling;
			}
		}
		for(var i = 0; i < sampleSet.length; i++){
			var mc = stage["swatch_"+i].holder;
			mc.clear();
			mc.lineStyle(1,0);
			mc.moveTo(0,0);
			var x = 0;
			var y = 0;
			for(var j = 0; j < sampleSet[i].length; j++){
				var theta = Math.PI * sampleSet[i][j];
				x += Math.cos(theta) * 3;
				y += Math.sin(theta) * 3;
				mc.lineTo(x, y);
			}
		}
	}
	// Fast random number list - based on Flash 5 code from:
	// http://www.the-stickman.com/tutorials/f5uniquerandom.shtml
	function randomIndex(total) {
		var numberlist = new Array();
		for (var i = 0; i < total; i++) {
			numberlist[i] = i;
		}
		var randomlist = new Array();
		for (var i = numberlist.length - 1; i > -1; i--) {
			var picked = Math.floor(Math.random()*i);
			randomlist[i] = numberlist[picked];
			numberlist[picked] = numberlist[i];
		}
    	return randomlist;
	}
}