﻿// Vector class for calculation of positions and directions

class Vector{
	var a:Object;// 	Old position x,y
	var b:Object;// 	Current position x,y
	var vx:Number;// 	distance between p0.x and p1.x
	var vy:Number;// 	distance between p0.y and p1.y
	var len:Number;// 	vector length (pythagoras length of p0 to p1)
	var dx:Number;// 	x value of unit vector (equal to Math.cos(theta))
	var dy:Number;// 	y value of unit vector (equal to Math.sin(theta))
	var rx:Number;// 	x value of right hand unit vector
	var ry:Number;// 	y value of right hand unit vector
	var lx:Number;// 	x value of left hand unit vector
	var ly:Number;// 	y value of left hand unit vector
	var theta:Number;// rotation - updated by calling atan2 or rotate methods
	// Empty constructor
	function Vector(a:Object, b:Object){
		this.a = a;
		this.b = b;
		updateVector();
	}
	// Initialise Vector
	function updateVector() {
		vx = b.x - a.x;
		vy = b.y - a.y;
		// length of vector
		len = Math.sqrt(vx * vx + vy * vy);
		// normalized unit-sized components
		if (len > 0) {
			dx = vx / len;
			dy = vy / len;
		} else {
			dx = 0.0;
			dy = 0.0;
		}
		// right hand normal
		rx = -dy;
		ry = dx;
		// left hand normal
		lx = dy;
		ly = -dx;
	}
	// Rotate vector around point A
	function rotateA(angle:Number){
		angle += atan2();
		theta = angle;
		b.x = Math.cos(angle) * len;
		b.y = Math.sin(angle) * len;
		updateVector();
	}
	// Rotate vector around point B
	function rotateB(angle:Number){
		angle += atan2();
		theta = angle;
		a.x = Math.cos(angle) * len;
		a.y = Math.sin(angle) * len;
		updateVector();
	}
	// Set theta and return it's value
	function atan2(){
		theta = Math.atan2(vy, vx);
		return theta;
	}
	// Perpendicular product (dot product rotated 90 degrees)
	static function perP(va, vb):Number{
		var pp = va.vx*vb.vy - va.vy*vb.vx;
		return pp;
	}
	// Mini perpendicular product (normalised dot product rotated 90 degrees)
	static function miniPerP(va, vb):Number{
		var pp = va.dx*vb.dy - va.dy*vb.dx;
		return pp;
	}
	// Returns a point on the same line as the surface you are projecting on to
	static function intersectionPoint(v1:Vector, v2:Vector):Object{
		var v3={vx:v2.a.x-v1.a.x, vy:v2.a.y-v1.a.y};
		var t=perP(v3, v2)/perP(v1, v2);
		var ip={};
		ip.x=v1.a.x+v1.vx*t;
		ip.y=v1.a.y+v1.vy*t;
		ip.t = t;
		return ip;
	}
	// Do two Vectors intersect?
	static function intersects(v1:Vector, v2:Vector):Boolean{
		var v3={vx:v2.a.x-v1.a.x, vy:v2.a.y-v1.a.y};
		var perp0 = perP(v3, v2);
		var perp1 = perP(v1, v2);
		var t0=perp0/perp1;
		var t1=perp1/perp0;
		return t0 >= 0 && t0 <= 1 && t1 >= 0 && t1 <= 1;
	}
	// Returns the dot product of two vectors
	static function dot(v1, v2):Number{
		return v1.vx*v2.vx + v1.vy*v2.vy;
	}
	// quick and dirty point to point math
	// Minus pB from pA and return a new object
	static function minus(pA, pB):Object{
		return {x:pA.x - pB.x, y:pA.y - pB.y};
	}
	// Add pB to pA and return a new object
	static function plus(pA, pB):Object{
		return {x:pA.x + pB.x, y:pA.y + pB.y};
	}
	// Return a new object of p scaled by num
	static function mul(p, num):Object{
		return {x:p.x * num, y:p.y * num};
	}
	// Return the distance between pA and pB
	static function dist(pA, pB):Number{
		return Math.sqrt((pB.x - pA.x) * (pB.x - pA.x) + (pB.y - pA.y) * (pB.y - pA.y));
	}
	// Return a point along the line pA to pB
	static function lerp(pA, pB, n:Number):Object{
		return {x:(pA.x + pB.x) * n, y:(pA.y + pB.y) * n};
	}
	// Return a normal of pA to pB
	static function normal(pA, pB):Object{
		var point = minus(pA, pB);
		var d = dist(pA, pB);
		point.x /= d;
		point.y /= d;
		return point;
	}
	// Return a copy of a point
	static function pCopy(p):Object{
		return {x:p.x, y:p.y};
	}
}